<?php

namespace Bydedi\Database\ActiveRecord;

use Bydedi\Database\ActiveRecord\Exception\ModelNotLoadedException;
use Bydedi\Database\ConnectionInterface;
use Bydedi\Database\Query;
use Exception;
use RuntimeException;

/**
 * Class Model
 *
 * @property \Bydedi\Database\ConnectionInterface $connection
 */
abstract class Model
{
    /**
     * @var \Bydedi\Database\ConnectionInterface
     */
    protected static $connection;
    /**
     * @var array
     */
    protected static $globalI18n = array();
    /**
     * @var array
     */
    protected static $gloabalPrimaryKey = array();
    /**
     * @var string
     */
    protected $suffixI18n = '_i18n';
    /**
     * @var bool
     */
    protected $i18n;
    /**
     * @var string
     */
    protected $primaryKey;
    /**
     * @var array
     */
    protected $innerJoin = array();
    /**
     * @var array
     */
    protected $leftJoin = array();
    /**
     * @var array
     */
    protected $where = array();
    /**
     * @var array
     */
    protected $extras = array();

    /**
     * @var array
     */
    protected $properties = array();

    /**
     * Constructeur
     *
     * @param integer|null $id id to load
     *
     * @throws \Exception
     */
    public function __construct($id = null)
    {

        if (empty($this->table)) {
            throw new Exception('properties $table is undefined');
        }

        $this->connection(self::$connection);

        $this->primaryKey = $this->resolvePrimaryKey();

        $this->i18n = $this->resolveI18nTable();

        if ($this->i18n) {
            $tableI18n = $this->table . $this->suffixI18n;
            $alias = Query::getAlias($this->table);
            $aliasI18n = Query::getAlias($tableI18n);

            $this->innerJoin[] = array(
                $tableI18n,
                'AND',
                array($alias . '.' . $this->primaryKey, $aliasI18n . '.' . $this->table . '_' . $this->primaryKey),
                array($aliasI18n . '.' . 'lang', '"' . getenv('APP_LOCALE') . '"')
            );
        }
        // --

        if (is_null($id) == false) {
            $this->load($id);
            if ($this->isLoaded() == false) {
                throw new ModelNotLoadedException(sprintf('the model %s can\'t load the id "%s"', get_class($this), $id));
            }
        }
    }

    /**
     * @return bool
     */
    public function resolveI18nTable()
    {
        if (!isset(static::$globalI18n[get_called_class()])) {
            static::$globalI18n[get_called_class()] = $this->getI18n();
        }

        return static::$globalI18n[get_called_class()];

    }

    /**
     * @return string
     */
    public function resolvePrimaryKey()
    {
        if (!isset(static::$gloabalPrimaryKey[get_called_class()])) {
            static::$gloabalPrimaryKey[get_called_class()] = $this->connection->getPrimaryKey($this->table);
        }

        return static::$gloabalPrimaryKey[get_called_class()];
    }

    /**
     * @param $id
     */
    public function load($id)
    {
        $this->hydrate($this->getOne(array(Query::getAlias($this->table) . '.' . $this->primaryKey => $id)));
    }

    /**
     * @param $data
     * @return bool
     */
    public function hydrate($data)
    {
        if (empty($data)) {
            return false;
        }

        foreach ($data as $k => $d) {
            $this->$k = $d;
        }

        return true;
    }

    /**
     * @param array $formvars
     * @param string $format
     *
     * @return array
     */
    public function get($formvars = array(), $format = null)
    {
        $results = $this->connection->getAll($this->getQuery($formvars)->build());

        if ($format == 'object') {
            $objects = [];

            foreach ($results as $result) {
                $obj = new static;
                $obj->hydrate($result);
                $objects[] = $obj;
            }

            return $objects;
        }

        return $results;
    }


    /**
     * @param array $formvars
     * @param string $format
     * @return array
     *
     * @throws Exception
     */
    public function getOne($formvars = array(), $format = null)
    {
        $result = $this->connection->getOne($this->getQuery($formvars)->build());

        if ($format == 'object') {
            $obj = new static;
            $obj->hydrate($result);
            return $obj;
        }

        return $result;
    }

    /**
     * @param $data
     * @return mixed
     * @throws \Exception
     *
     */
    public function save($data = null)
    {
        if (property_exists($this, 'table') == false) {
            throw new Exception(
                'No property table in class ' . get_class($this)
            );
        }

        if (empty($this->primaryKey)) {
            throw new Exception(
                'No primary key defined in table ' . $this->table
            );
        }

        if (is_null($data)) {
            $data = $this->properties;
        }

        $id = $this->connection->save($this->table, $data);

        if (empty($id)) {
            throw new RuntimeException('An error occur when saving data.');
        }

        if ($this->isI18n() && isset($data["lang"])) {
            $data[$this->table . "_id"] = $id;
            $this->connection->save($this->table . $this->suffixI18n, $data);
        }

        return $id;
    }

    /**
     * @return bool
     */
    public function isI18n()
    {
        return $this->i18n;
    }

    /**
     * @return bool
     *
     * check si table $this->table . "_nom" exist
     */
    private function getI18n()
    {
        $sql = "SHOW TABLES LIKE '" . $this->table . $this->suffixI18n . "'";
        $result = $this->connection->getAll($sql);

        return (!empty($result));
    }

    /**
     * @param array $formvars
     * @return Query()
     *
     * build request with table, i18n,
     * property innerJoin / leftJoin, where and options in $formvars
     */
    public function getQuery($formvars = array())
    {

        $q = new Query();
        $q->select()
            ->from($this->table);

        if (!empty($this->innerJoin)) {
            foreach ($this->innerJoin as $ij) {
                $q->join('innerJoin', $ij);
            }
        }

        if (!empty($this->leftJoin)) {
            foreach ($this->leftJoin as $lj) {
                $q->join('leftJoin', $lj);
            }
        }

        if (!empty($this->where)) {
            foreach ($this->where as $w) {
                $q->where($w);
            }
        }

        $formvars = array_merge($this->extras, $formvars);

        if (!empty($formvars)) {
            foreach ($formvars as $k => $v) {
                if (method_exists($q, $k)) {
                    $q->$k($v);
                } else {
                    $q->where($k, $v);
                }
            }
        }

        return $q;
    }

    /**
     * Apply list of transformer on the data result
     *
     * @param $data
     * @return int coutn transformers apply
     */
    public function applyTranformers(&$data)
    {
        $apply = 0;
        if (!empty(static::$transformers)) {
            foreach (static::$transformers as $transformer) {
                if (!in_array('Transformer\\TransformerInterface', class_implements($transformer))) {
                    continue;
                }

                if ($this->array_depth($data) > 1) {
                    foreach ($data as $k => $d) {
                        $data[$k] = call_user_func_array(array(new $transformer, "transform"), array($d));
                    }
                } else {
                    $data = call_user_func_array(array(new $transformer, "transform"), $data);
                }
                $apply++;
            }
        }

        return $apply;
    }

    /**
     * Return properties in array
     *
     * @return array
     */
    public function toArray()
    {
        return $this->properties;
    }

    /**
     * get the db connection
     *
     * @param ConnectionInterface|null $newConnection
     *
     * @return ConnectionInterface
     */
    protected function connection(ConnectionInterface $newConnection = null)
    {
        static $connection;
        if ($newConnection !== null) {
            $connection = $newConnection;
        }

        return $connection;
    }

    /**
     * @param $name
     * @return mixed
     */
    public function __get($name)
    {

        if ($name == 'connection') {
            return $this->connection();
        }

        if (array_key_exists($name, $this->properties)) {
            return $this->properties[$name];
        }
    }

    /**
     * @param $name
     * @param $value
     */
    public function __set($name, $value)
    {
        $this->properties[$name] = $value;
    }

    /**
     * @param $name
     * @return bool
     */
    public function __isset($name)
    {
        return !empty($this->properties[$name]);
    }

    /**
     * @param array $array
     * @return int
     *
     * return the depth of the array pass in parameter
     */
    protected function array_depth(array $array)
    {
        $max_depth = 1;

        foreach ($array as $value) {
            if (is_array($value)) {
                $depth = $this->array_depth($value) + 1;

                if ($depth > $max_depth) {
                    $max_depth = $depth;
                }
            }
        }

        return $max_depth;
    }

    public function isLoaded()
    {
        return !empty($this->properties[$this->primaryKey]);
    }

    public static function setConnection(ConnectionInterface $connection)
    {
        self::$connection = $connection;
    }
}
